@@ -11,7 +11,7 @@ from .server import *
 
 @asyncio.coroutine
 def echo(ws, uri):
-    ws.send((yield from ws.recv()))
+    yield from ws.send((yield from ws.recv()))
 
 
 class ClientServerTests(unittest.TestCase):
@@ -42,7 +42,7 @@ class ClientServerTests(unittest.TestCase):
 
     def test_basic(self):
         self.start_client()
-        self.client.send("Hello!")
+        self.loop.run_until_complete(self.client.send("Hello!"))
         reply = self.loop.run_until_complete(self.client.recv())
         self.assertEqual(reply, "Hello!")
         self.stop_client()
@@ -61,6 +61,10 @@ class ClientServerTests(unittest.TestCase):
         with self.assertRaises(InvalidHandshake):
             self.start_client()
 
+        # Now the server believes the connection is open. Run the event loop
+        # once to make it notice the connection was closed. Interesting hack.
+        yield from tulip.sleep(0)
+
     @patch('websockets.client.build_request')
     def test_client_sends_invalid_handshake_request(self, _build_request):
         def wrong_build_request(set_header):
@@ -95,7 +99,7 @@ class ClientServerTests(unittest.TestCase):
         send.side_effect = ValueError("send failed")
 
         self.start_client()
-        self.client.send("Hello!")
+        self.loop.run_until_complete(self.client.send("Hello!"))
         reply = self.loop.run_until_complete(self.client.recv())
         self.assertEqual(reply, None)
         self.stop_client()
@@ -108,7 +112,7 @@ class ClientServerTests(unittest.TestCase):
         close.side_effect = ValueError("close failed")
 
         self.start_client()
-        self.client.send("Hello!")
+        self.loop.run_until_complete(self.client.send("Hello!"))
         reply = self.loop.run_until_complete(self.client.recv())
         self.assertEqual(reply, "Hello!")
         self.stop_client()
